/*
 * http - Firestorm HTTP decode plugin
 * Copyright (c) 2002 Gianni Tedesco <gianni@scaramanga.co.uk>
 * Released under the terms of the GNU GPL v2
 *
 * TODO:
 *  o Add data layers
 *  o Put CONNECT support back in
 *  o Support chunked transfer encoding
 *  o Rip out path part of URL
 *  o Merge URL normalization code in
 *  o Need to test pipelining properly
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <firestorm.h>
#include <args.h>
#include <packet.h>
#include <plugin.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>
#include <netinet/in.h>

PLUGIN_STD_DEFS();
static proc_dispatch dispatch;
extern struct proto http_p;

/* Accepted HTTP versions, must correspond with the  */
struct {
	int maj, min;
}http_proto_vers[HTTP_VER_MAX]={
	[HTTP_VER_UNKNOWN] {-1, -1},
	[HTTP_VER_0_9] {0, 9},
	[HTTP_VER_1_0] {1, 0},
	[HTTP_VER_1_1] {1, 1},
};

/* HTTP Decode Control Block */
struct http_dcb {
	const char *label;
	void (*fn)(struct http_dcb *, struct fvec *);
	union {
		struct fvec *vec;
		int *val;
	}u;
};

/* Get a version code for a given major and minor version */
static inline unsigned int http_vers_idx(int maj, int min)
{
	int i;
	
	for(i=1; i < HTTP_VER_MAX; i++) {
		if ( http_proto_vers[i].maj == maj && 
			http_proto_vers[i].min == min )
		return i;
	}
	
	return HTTP_VER_UNKNOWN;
}

/* Parse and HTTP version string (eg: "HTTP/1.0") */
int http_proto_version(struct fvec *str)
{
	const char *s = str->ptr;
	int maj, min;
	int ret;
	
	if ( str->ptr == NULL )
		return HTTP_VER_UNKNOWN;

	if ( str->len != 8 ) /* sizeof("HTTP/X.Y") */
		return HTTP_VER_UNKNOWN;
	
	if ( memcmp(s, "HTTP/", 5) )
		return HTTP_VER_UNKNOWN;
	
	if ( s[6] != '.' )
		return HTTP_VER_UNKNOWN;

	if ( !isdigit(s[5]) || !isdigit(s[7]) )
		return HTTP_VER_UNKNOWN;
	
	maj = s[5] - '0';
	min = s[7] - '0';
	ret = http_vers_idx(maj, min);

	return ret;
}

static void htype_string(struct http_dcb *h, struct fvec *v)
{
	if ( !h->u.vec )
		return;

	h->u.vec->ptr = v->ptr;
	h->u.vec->len = v->len;
}

static void htype_int(struct http_dcb *h, struct fvec *v)
{
	unsigned int val;
	int len;

	len = vtouint(v, &val);
	if ( len > 0 ) {
		*h->u.val = val;
	}else{
		*h->u.val = val;
	}
}

/* Same as int but ensure input is only 3 digits */
static void htype_code(struct http_dcb *h, struct fvec *v)
{
	if ( v->len != 3 ) {
		*h->u.val = -1;
		return;
	}

	htype_int(h, v);
}

/* Constructor for HTTP state */
static void http_ctor(void *priv)
{
	struct http_flow *f=priv;
	f->client.state = HTTPS_HEADER;
	f->server.state = HTTPS_HEADER;
}

/* Hex decode a single ASCII char */
static inline int hexchar(unsigned int c)
{
	if ( c - '0' < 10 )
		return c - '0';
	if ( c - 'a' < 6 )
		return (c - 'a') + 10;
	if ( c - 'A' < 6 )
		return (c - 'A') + 10;

	return -1;
}

static int http_dprint(struct layer *l, char *buf, int buflen)
{
	struct http_request *h;
	char *obuf=buf;

	if ( !(h=l->session) )
		return snprintf(buf, buflen, "[No session info]");

	/* 27 = 13+12+2 = bits of strings */
	if ( h->type == HTTP_REQUEST ) {
		if ( buflen < 27 + h->uri.len + h->method.len )
			return snprintf(buf, buflen, "buffer too small");

		memcpy(buf, "http_method='", 13); buf+=13;
		memcpy(buf, h->method.ptr, h->method.len); buf+=h->method.len;
		memcpy(buf, "' http_uri='", 12); buf+=12;
		memcpy(buf, h->uri.ptr, h->uri.len); buf+=h->uri.len;
		memcpy(buf, "'", 2); /* including trailing null */
	}else{
		struct http_response *r = l->session;
		snprintf(buf, buflen, "http_code='%i'", r->code);
	}

	return (buf-obuf);
}

static int http_serialize(struct packet *p, unsigned int i, char *buf, int len)
{
	struct layer *l=&p->layer[i];
	struct http_request *s=(struct http_request *)l->session;
	struct http_response *sr=(struct http_response *)l->session;
	struct http_serial_req *f=(struct http_serial_req *)buf;
	struct http_serial_resp *r=(struct http_serial_resp *)buf;

	if ( len < sizeof(struct http_serial_req) )
		return -1;

	if ( s->type == HTTP_REQUEST ) {
		f->type = s->type;
		f->vers = s->proto_vers;
		f->unused = 0;
		f->uri = htonl(s->uri.ptr - l->h.raw);
		f->uri_len = htonl(s->uri.len);
		f->method = htonl(s->method.ptr - l->h.raw);
		f->method_len = htonl(s->method.len);
		return sizeof(*f);
	}else{
		r->type = sr->type;
		r->vers = sr->proto_vers;
		r->unused = 0;
		r->code = htonl(sr->code);
		return sizeof(*f);
	}
}

/* For backwards compatibility */
static void *http_deserialize_old(struct packet *p, unsigned int i,
				char *buf, int len)
{
	struct http_request *s;
	struct http_serial *f = (struct http_serial *)buf;
	struct layer *l=&p->layer[i];

	s = calloc(1, sizeof(*s));
	if ( s == NULL )
		return NULL;

	s->uri.ptr = l->h.raw + ntohl(f->uri);
	s->uri.len = ntohl(f->uri_len);
	s->method.ptr = l->h.raw + ntohl(f->method);
	s->method.len = ntohl(f->method_len);
	return s;
}

/* Remember: this data is un-trusted so validate it! */
static void *http_deserialize(struct packet *p, unsigned int i,
				char *buf, int len)
{
	struct http_request *s;
	struct http_response *sr;
	struct http_serial_req *f=(struct http_serial_req *)buf;
	struct http_serial_resp *r=(struct http_serial_resp *)buf;
	struct layer *l=&p->layer[i];
	u_int32_t plen;

	plen=(char *)p->end - (char *)l->h.raw;
	
	if ( len == sizeof(struct http_serial) )
		return http_deserialize_old(p, i, buf, len);

	if ( len != sizeof(*f) && len != sizeof(*r) )
		return NULL;
	
	if ( f->unused != 0 )
		return NULL;

	switch ( f->type ) {
	case HTTP_REQUEST:
		if ( !(s=calloc(1, sizeof(*s))) )
			return NULL;

		if ( plen < ntohl(f->uri)+ntohl(f->uri_len) )
			goto fail;
		if ( plen < ntohl(f->method)+ntohl(f->method_len) )
			goto fail;

		s->type = f->type;
		s->proto_vers = f->vers;
		s->uri.ptr = l->h.raw + ntohl(f->uri);
		s->uri.len = ntohl(f->uri_len);
		s->method.ptr = l->h.raw + ntohl(f->method);
		s->method.len = ntohl(f->method_len);

		return s;

	case HTTP_RESPONSE:
		if ( !(s=calloc(1, sizeof(*sr))) )
			return NULL;

		sr = (struct http_response *)s;
		sr->type = r->type;
		sr->proto_vers = r->vers;
		sr->code = (int)ntohl(r->code);

		return sr;
	default:
		return NULL;
	}

fail:
	free(s);
	return NULL;
}

/* Check if this header is one we want to store */
static inline void dispatch_hdr(struct http_dcb *d,
				struct fvec *k,
				struct fvec *v)
{
	for(; d->label; d++) {
		if ( vstrcmp(k, d->label) )
			continue;
		d->fn(d, v);
	}
}

/* Actually parse an HTTP request */
static size_t http_decode_buf(struct http_dcb *d, char *p, char *end)
{
	char *cur;
	struct fvec hv[3]; /* method, url, proto */
	struct fvec k,v;
	int i=0;
	int state=0;
	int ret=0;

	hv[0].ptr=NULL;
	hv[1].ptr=NULL;
	hv[2].ptr=NULL;

	for(cur=p; cur<end; cur++) {
		switch ( state ) {
		case 0:
			if ( *cur != ' ' ) {
				state=1;
				hv[i].ptr=cur;
				hv[i].len=0;
			}
			break;
		case 1:
			hv[i].len++;
			switch(*cur) {
			case ' ':
				if ( i<2 ) {
					state=0;
					i++;
				}
				break;
			case '\n':
				if ( hv[i].len && *(cur-1)=='\r' )
					hv[i].len--;
				k.ptr = cur + 1;
				k.len = 0;
				state = 2;
				break;
			}
			break;
		case 2:
			if ( *cur == ':' ) {
				state = 3;
				break;
			}else if ( *cur=='\n' ) {
				ret = (cur - p) + 1;
				cur = end;
			}
			k.len++;
			break;
		case 3:
			if ( *cur != ' ' ) {
				v.ptr = cur;
				v.len = 0;
				state = 4;
			}
			break;
		case 4:
			v.len++;
			if ( *cur == '\n' ) {
				if ( v.len && *(cur-1)=='\r' ) {
					v.len--;
				}
				dispatch_hdr(d+3, &k, &v);
				k.ptr = cur + 1;
				k.len = 0;
				state = 2;
				break;
			}
			break;
		}
	}

	if ( hv[0].ptr==NULL || hv[1].ptr==NULL )
		return 0;

	/* Setup method/url/proto */
	d[0].fn(&d[0], &hv[0]);
	d[1].fn(&d[1], &hv[1]);
	d[2].fn(&d[2], &hv[2]);

	return ret;
}

/* Return 0 if we need to start reassembling */
static size_t http_sdecode(struct packet *p, char *buf, size_t len)
{
	struct tcp_session *s;
	struct http_fside f;
	char *begin, *end;
	int req=1;

	begin = buf;
	end = buf + len;

	/* Get the state information */
	s = p->layer[p->llen-1].session;
	if ( s->flow ) {
		struct http_flow *flow = s->flow;
		if ( p->layer[p->llen-1].flags & FLAG_TCP_2SVR ) {
			f.state = flow->client.state;
			f.len = flow->client.len;
		}else{
			f.state = flow->server.state;
			f.len = flow->server.len;
			req = 0;
		}
	}else{
		f.state = HTTPS_HEADER;
		if ( p->layer[p->llen-1].flags & FLAG_TCP_2SVR)
			req = 0;
	}

	/* Parse as much data as possible */
again:
	if ( f.state == HTTPS_HEADER ) {
		size_t hlen;
		int clen=-1;
		struct http_dcb dcb[]={
			{"method", htype_string, NULL},
			{"uri", htype_string, NULL},
			{"protocol", htype_string, NULL},
			{"Content-Length", htype_int, {.val=&clen}},
			{NULL,}
		};

		hlen = http_decode_buf(dcb, begin, end);
		if ( !hlen )
			return begin - buf;

		begin += hlen;
		
		if ( clen > 0 ) {
			f.state = HTTPS_CONTENT;
			f.len = clen;
		}else if ( !req && clen<0 ){
			f.state = HTTPS_CLOSING;
			return len;
		}
	}else if ( f.state == HTTPS_CONTENT ) {
		size_t rlen = end-begin;

		if ( rlen >= f.len ) {
			begin += f.len;
			f.len = 0;
			f.state = HTTPS_HEADER;
		}else{
			f.len -= rlen;
			begin += rlen;
		}
	}else if ( f.state == HTTPS_CLOSING ) {
		return len;
	}else{
		mesg(M_DEBUG, "Aaaargh");
		return len;
	}

	if ( begin < end )
		goto again;

	if ( begin > end )
		begin = end;

	return begin - buf;
}

/* Parse and pump an HTTP request */
size_t http_request(struct packet *p, struct http_fside *f,
			char *begin, char *end)
{
	int clen=-1;
	size_t hlen;
	struct http_request r;
	struct fvec pv = {.ptr = NULL};
	struct http_dcb dcb[]={
		{"method", htype_string, {.vec = &r.method}},
		{"uri", htype_string, {.vec = &r.uri}},
		{"protocol", htype_string, {.vec = &pv}},
		{"Content-Length", htype_int, {.val = &clen}},
		{NULL,}
	};

	hlen = http_decode_buf(dcb, begin, end);
	if ( !hlen )
		return 0;

	p->layer[p->llen].session = &r;
	p->llen++;

	r.type = HTTP_REQUEST;
	r.proto_vers = http_proto_version(&pv);

	if ( clen > 0 ) {
		f->state = HTTPS_CONTENT;
		f->len = clen;
	}

	if ( !vstrcmp(&r.method, "CONNECT") ) {
		/* TODO: Implement */
	}

	dispatch(p);
	return hlen;
}

/* Parse and pump an HTTP response */
size_t http_response(struct packet *p, struct http_fside *f,
			char *begin, char *end)
{
	int clen=-1;
	size_t hlen;
	struct http_response r;
	struct fvec pv = {.ptr = NULL};
	struct http_dcb dcb[]={
		{"protocol", htype_string, {.vec = &pv}},
		{"code", htype_code, {.val = &r.code}},
		{"msg", htype_string, NULL},
		{"Content-Length", htype_int, {.val = &clen}},
		{NULL,}
	};

	hlen = http_decode_buf(dcb, begin, end);
	if ( !hlen )
		return 0;

	p->layer[p->llen].session = &r;
	p->llen++;

	r.type = HTTP_RESPONSE;
	r.proto_vers = http_proto_version(&pv);

	if ( clen > 0 ) {
		f->state = HTTPS_CONTENT;
		f->len = clen;
	}else if ( clen < 0 ) {
		f->state = HTTPS_CLOSING;
	}

	dispatch(p);
	return hlen;
}

/* HTTP decode callback */
static void http_decode(struct packet *p)
{
	struct http_flow noflow;
	struct http_flow *flow;
	struct tcp_session *s;
	struct http_fside *f;
	char *begin, *end;
	int olen = p->llen;
	int dispatches=0;
	size_t (*fn)(struct packet *, struct http_fside *, char *, char *);

	/* Grab the data portion */
	begin = p->layer[p->llen].h.raw;
	end = p->layer[p->llen-2].h.raw +
		ntohs(p->layer[p->llen-2].h.ip->tot_len);

	if ( begin >= end )
		goto done;

	/* Setup state information */
	s = p->layer[p->llen-1].session;
	if ( s->flow ) {
		flow = s->flow;
	}else{
		flow = &noflow;
		http_ctor(flow);
	}

	/* Get the right side */
	if ( p->layer[p->llen-1].flags & FLAG_TCP_2SVR ) {
		f = &flow->client;
		fn = http_request;
	}else{
		f = &flow->server;
		fn = http_response;
	}

again:
	if ( f->state == HTTPS_HEADER ) {
		size_t hlen;

		hlen = fn(p, f, begin, end);
		if ( !hlen )
			goto done;

		dispatches++;
		p->llen = olen;

		begin += hlen;
	}else if ( f->state == HTTPS_CONTENT ) {
		size_t rlen = end-begin;

		if ( rlen >= f->len ) {
			begin += f->len;
			f->len = 0;
			f->state = HTTPS_HEADER;
		}else{
			f->len -= rlen;
			begin += rlen;
		}
	}else if ( f->state == 3) {
		goto done;
	}else{
		mesg(M_DEBUG, "corrupt flow");
	}

	if ( begin < end )
		goto again;

done:
	if ( !dispatches )
		dispatch(p);
}

struct proto http_p={
	.name="http",
	.sdecode=http_sdecode,
	.decode=http_decode,
	.print=http_dprint,
	.flow_len=sizeof(struct http_flow),
	.flow_ctor=http_ctor,
	.serialize=http_serialize,
	.deserialize=http_deserialize,
	.free=free,
};

struct proto_req http_r[]={
	proto_request("tcp", __constant_htons(80)),
	proto_request("tcp", __constant_htons(3128)),
	proto_request("tcp", __constant_htons(8080)),
	null_request()
};

int PLUGIN_DECODE (struct decode_api *d)
{
	int ok=0;

	object_check(d);

	dispatch=d->dispatch;

	ok+=d->decode_add(&http_p, http_r);

	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	plugin_check(in, out);

	PLUGIN_ID("decode.http", "Hyper-text Transfer Protocol");
	PLUGIN_VERSION(0, 1);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@scaramanga.co.uk");
	PLUGIN_LICENSE("GPL");

	return PLUGIN_ERR_OK;
}

int PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
