/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is the detection engine, it is generic to all protocols
* although each protocol is expected to maintain its own root
* node. It is a very fast binary tree like structure. If you
* want more details mail me ;) -- GT
*/
#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <firestorm.h>
#include <packet.h>
#include <cleanup.h>
#include <alert.h>
#include <signature.h>
#include <matcher.h>
#include <detect.h>

static struct alert *cur_alert;

/* Sort rules by cost */
static int m_sorter(const void *p1, const void *p2)
{
	const struct sig_node *r1=p1;
	const struct sig_node *r2=p2;

	if ( r1->cost < r2->cost) return -1;
	if ( r1->cost > r2->cost) return 1;

	if ( !r1->n && r2->n ) return -1;
	if ( r1->n && !r2->n ) return 1;

	/* Tertiary sorting, based on function,
	 * pointer, just to be consistent */
	if ( r1->m < r2->m ) return -1;
	if ( r1->m > r2->m ) return 1;

	return 0;
}

/* Run a packet against a tree */
static void do_detect(struct sig_node *n, struct packet *p, unsigned int l)
{
	for(; n; n=n->next) {
		if ( n->m(p, n->p, l, n->n) ) {
			if ( n->a ) {
				if ( !cur_alert ) {
					cur_alert=n->a;
				}else if ( cur_alert->nr > n->a->nr ) {
					cur_alert=n->a;
				}
			}
			do_detect(n->child, p, l);
		}
	}
}

void detect(struct sig_node *n, struct packet *p, unsigned int l)
{
	cur_alert=NULL;

	if ( n == NULL )
		return;

	do_detect(n->child, p, l);

	if ( cur_alert )
		alert(cur_alert, p);
}

static void do_optimize(struct sig_node *n)
{
	if ( n->child )
		do_optimize(n->child);

	if ( n->match->optimize )
		n->match->optimize(n);
	
	if ( n->next )
		do_optimize(n->next);
}

void detect_optimize(struct sig_node *r)
{
	if ( r->child )
		do_optimize(r->child);
}

/* Add a signature to a tree */
int detect_add_sig(struct sig_node *x, unsigned int num,
		struct sig_node *cur, struct alert *a)
{
	struct sig_node *n, *use=NULL;
	unsigned int i;
	int t;

	/* Sort the list by cost */
	qsort(x, num, sizeof(*x), m_sorter);

	/* The last item is the alert */
	x[num-1].a=a;

	/* Add the nodes to the tree  */
	for(i=0; i<num; i++) {
		/* We may need to go above everyone else if some
		 * higher cost nodes were added first */
		if (cur->child && m_sorter(cur->child, &x[i])>0) {
			x[i].next=cur->child;
			cur->child=&x[i];
			cur=&x[i];
			continue;
		}

		/* Skip past lower cost nodes than us that are
		 * not relevent to this rule */
		if ( cur->child && m_sorter(cur->child, &x[i])<0 ) {
			struct sig_node *tmp;

			/* There is already a child but it is higher
			 * priority than us, so skip past all higher
			 * priority nodes on its next pointer */
			for(tmp=cur->child; tmp; tmp=tmp->next) {
				if ( m_sorter(tmp, &x[i])>=0 )
					break;
				use=tmp;
			}
			t=1;
		}else{
			/* If there is a child, it is lower priority
			 * than us, so we take its spot and it can hang
			 * off of our next pointer */
			use=cur->child;
			t=2;
		}

		/* Check if an identical node already exists in the
		 * tree, if so then we can use that instead */
		for(n=use; n; n=n->next) {
			if ( t==1 && m_sorter(n, &x[i])>0 )
				break;

			/* Check if node can be absorbed */
			if ( n->match->absorb ) {
				if ( n->match->absorb(n, &x[i]) ) {
					/* Make rest of the nodes hang off the end */
					for(; i<num-1; i++)
						x[i].child=&x[i+1];
					return 1;
				}
			}

			/* Then check if nodes are identical */
			if ( n->n==x[i].n && n->m==x[i].m &&
				n->match->compare(n->p, x[i].p)==CMP_EQUAL ) {

				/* Make sure we don't lose the alert by
				 * forgetting about this node just
				 * because a similar one exists */
				if ( x[i].a )
					n->a=x[i].a;

				if ( x[i].match->cleanup )
					x[i].match->cleanup(x[i].p);

				cur=n;
				t=0;
				break;
			}
		}

		/* Destroy this node and go on to the next,
		 * if an identical node was found */
		if ( t==0 )
			continue;

		/* Add ourselves to the bottom of the tree */
		if ( t==2 ) {
			x[i].next=cur->child;
			cur->child=&x[i];
			cur=&x[i];
		}else{
			x[i].next=use->next;
			use->next=&x[i];
			cur=&x[i];
		}
	}

	return 1;
}

/* Delete a signature, if bits of it are in the tree,
 * this will fuckup the tree. Eventually this code will
 * allow 'hot-plugging' and 'unplugging' of signatures. */
void detect_free_sig(struct sig_node *r, unsigned int n)
{
	unsigned int i;

	if (!r) return;

	for(i=0; i<n; i++) {
		r[i].match->cleanup(r[i].p);
	}

	free(r);
}
