/*
* This file is part of tcpidx
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*
* This code takes a bunch of i_ordered databses and squeezes them all
* out in to one big database file. On-disk formats are described in
* idx.h.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <firestorm.h>
#include <packet.h>
#include <alert.h>

#include <mpool.h>
#include <fbuf.h>
#include <idx.h>
#include <i_ordered.h>
#include <i_idx.h>

struct i_idx *i_idx_new(int num_fields)
{
	struct i_idx *ret;
	size_t len;

	len = sizeof(*ret) + num_fields * sizeof(struct i_field);
	ret = malloc(len);

	if ( !ret )
		return NULL;

	ret->num_fields = num_fields;
	ret->cur_field = 0;

	return ret;
}

int i_idx_ordered_field(struct i_idx *idx,
			const char *name,
			struct i_ordered *db)
{
	if ( !name || !db )
		return 0;

	if ( idx->cur_field >= idx->num_fields )
		return 0;

	idx->field[idx->cur_field].name = name;
	idx->field[idx->cur_field].db = db;
	idx->cur_field++;

	return 1;
}

int i_idx_write_fd(struct i_idx *idx, int fd)
{
	struct fbuf o;
	struct idx_hdr hdr;
	idx_ofs_t ofs;
	int ret = 1;
	int i;

	if ( fbuf_attach(&o, fd, 32768)<0 )
		return 0;

	/* write header */
	hdr.magic = IDX_MAGIC;
	hdr.num = idx->cur_field;
	fbuf_write(&o, &hdr, sizeof(hdr));

	ofs = sizeof(hdr) + hdr.num * sizeof(struct idx_index);

	/* do the writes */
	for(i=0; i<idx->cur_field; i++) {
		struct idx_index db;
		struct i_field *f=&idx->field[i];

		strncpy(db.name, f->name, sizeof(db.name));
		db.type = IDX_TYPE_ORDERED;
		db.reserved1 = 0;
		db.reserved2 = 0;
		db.ofs = ofs;
		db.len = i_ordered_size(f->db);
		db.num_keys = i_ordered_keys(f->db);
		db.num_vals = i_ordered_vals(f->db);

		fbuf_write(&o, &db, sizeof(db));
		ofs += db.len;
	}
	
	for(i=0; i<idx->cur_field; i++) {
		struct i_field *f=&idx->field[i];

		if ( i_ordered_write(f->db, &o)<0 ) {
			ret = 0;
			break;
		}

		i_ordered_free(f->db);
	}

	if ( i < idx->cur_field) {
		for(; i < idx->cur_field; i++)
			i_ordered_free(idx->field[i].db);
	}

	if ( fbuf_close(&o) < 0 )
		ret = 0;

	return ret;
}

void i_idx_free(struct i_idx *idx)
{
	free(idx);
}
